from __future__ import annotations

import argparse
import json
import shutil
import subprocess
import sys
import tempfile
import webbrowser
from pathlib import Path

import nox
from nox import Session

from exasol.toolbox.config import BaseConfig
from exasol.toolbox.nox._shared import DOCS_OUTPUT_DIR
from noxconfig import (
    PROJECT_CONFIG,
)


def _build_docs(session: nox.Session, config: BaseConfig) -> None:
    session.run(
        "sphinx-build",
        "-W",
        "-b",
        "html",
        f"{config.documentation_path}",
        DOCS_OUTPUT_DIR,
    )


def _build_multiversion_docs(session: nox.Session, config: BaseConfig) -> None:
    session.run(
        "sphinx-multiversion",
        f"{config.documentation_path}",
        DOCS_OUTPUT_DIR,
    )
    session.run("touch", f"{DOCS_OUTPUT_DIR}/.nojekyll")


def _git_diff_changes_main() -> int:
    """
    Check if doc/changes is changed and return the exit code of command git diff.
    The exit code is 0 if there are no changes.
    """
    p = subprocess.run(
        [
            "git",
            "diff",
            "--quiet",
            "origin/main",
            "--",
            PROJECT_CONFIG.documentation_path / "changes",
        ],
        capture_output=True,
    )
    return p.returncode


@nox.session(name="docs:multiversion", python=False)
def build_multiversion(session: Session) -> None:
    """Builds the multiversion project documentation"""
    _build_multiversion_docs(session, PROJECT_CONFIG)


@nox.session(name="docs:build", python=False)
def build_docs(session: Session) -> None:
    """Builds the project documentation"""
    _build_docs(session, PROJECT_CONFIG)


@nox.session(name="docs:open", python=False)
def open_docs(session: Session) -> None:
    """Opens the built project documentation"""
    docs_folder = PROJECT_CONFIG.root_path / DOCS_OUTPUT_DIR
    if not docs_folder.exists():
        session.error(f"No documentation could be found. {docs_folder} is missing")
    index = docs_folder / "index.html"
    webbrowser.open_new_tab(index.as_uri())


@nox.session(name="docs:clean", python=False)
def clean_docs(_session: Session) -> None:
    """Removes the documentations build folder"""
    docs_folder = PROJECT_CONFIG.root_path / DOCS_OUTPUT_DIR
    if docs_folder.exists():
        shutil.rmtree(docs_folder)


def _docs_list_links(doc_config: Path):
    with tempfile.TemporaryDirectory() as path:
        tmpdir = Path(path)
        sp = subprocess.run(  # nosec
            [
                "sphinx-build",
                "-b",
                "linkcheck",
                "-D",
                "linkcheck_ignore=.*",
                doc_config,
                tmpdir,
            ],
            capture_output=True,
            text=True,
        )
        if sp.returncode >= 2:
            return sp.returncode, sp.stderr
        output = tmpdir / "output.json"
        links = output.read_text().split("\n")
        file_links = []
        for link in links:
            if link != "":
                line = json.loads(link)
                if not line["uri"].startswith("#"):
                    file_links.append(line)
        file_links.sort(key=lambda file: file["filename"])
        return 0, "\n".join(
            f"filename: {fl['filename']}:{fl['lineno']} -> uri: {fl['uri']}"
            for fl in file_links
        )


def _docs_links_check(doc_config: Path, args):
    with tempfile.TemporaryDirectory() as path:
        tmpdir = Path(path)
        sp = subprocess.run(  # nosec
            [
                "sphinx-build",
                "-b",
                "linkcheck",
                doc_config,
                tmpdir,
            ],
        )
        if args.output and sp.returncode <= 1:
            result_json = tmpdir / "output.json"
            dst = Path(args.output) / "link-check-output.json"
            shutil.copyfile(result_json, dst)
            print(f"file generated at path: {result_json.resolve()}")
        return sp.returncode, (
            None if sp.returncode >= 2 else (tmpdir / "output.txt").read_text()
        )


@nox.session(name="links:list", python=False)
def docs_list_links(session: Session) -> None:
    """List all the links within the documentation."""
    r_code, text = _docs_list_links(PROJECT_CONFIG.documentation_path)
    print(text)
    if r_code != 0:
        session.error()


@nox.session(name="links:check", python=False)
def docs_links_check(session: Session) -> None:
    """Checks whether all links in the documentation are accessible."""
    parser = argparse.ArgumentParser(
        prog="nox -s links:check",
        usage="nox -s links:check -- [-h] [-o |--output]",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-o", "--output", type=Path, help="path to copy the output json", default=None
    )
    args = parser.parse_args(session.posargs)
    r_code, problems = _docs_links_check(PROJECT_CONFIG.documentation_path, args)
    if r_code >= 2:
        session.error(2)
    if r_code == 1 or problems != "":
        escape_red = "\033[31m"
        print(escape_red + "errors:")
        print(problems)
        session.error(1)


@nox.session(name="changelog:updated", python=False)
def updated(_session: Session) -> None:
    """Checks if the change log has been updated"""
    if _git_diff_changes_main() == 0:
        print(
            "Changelog is not updated.\n"
            "Please describe your changes in the changelog!"
        )
        sys.exit(1)



from pydantic import (
    BaseModel,
    ConfigDict,
)

from exasol.toolbox.util.dependencies.audit import Vulnerability


class ResolvedVulnerabilities(BaseModel):
    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    previous_vulnerabilities: list[Vulnerability]
    current_vulnerabilities: list[Vulnerability]

    def _is_resolved(self, previous_vuln: Vulnerability):
        """
        Detects if a vulnerability has been resolved.

        A vulnerability is said to be resolved when it cannot be found
        in the `current_vulnerabilities`. In order to see if a vulnerability
        is still present, its id and aliases are compared to values in the
        `current_vulnerabilities`. It is hoped that if an ID were to change
        that this would still be present in the aliases.
        """
        previous_vuln_set = {previous_vuln.id, *previous_vuln.aliases}
        for current_vuln in self.current_vulnerabilities:
            if previous_vuln.package.name == current_vuln.package.name:
                current_vuln_id_set = {current_vuln.id, *current_vuln.aliases}
                if previous_vuln_set.intersection(current_vuln_id_set):
                    return False
        return True

    @property
    def resolutions(self) -> list[Vulnerability]:
        """
        Return resolved vulnerabilities
        """
        resolved_vulnerabilities = []
        for previous_vuln in self.previous_vulnerabilities:
            if self._is_resolved(previous_vuln):
                resolved_vulnerabilities.append(previous_vuln)
        return resolved_vulnerabilities



from __future__ import annotations

import json
import subprocess  # nosec
import tempfile
from dataclasses import dataclass
from enum import Enum
from inspect import cleandoc
from pathlib import Path
from re import search
from typing import (
    Any,
)

from pydantic import (
    BaseModel,
    ConfigDict,
)

from exasol.toolbox.util.dependencies.shared_models import (
    Package,
    poetry_files_from_latest_tag,
)

PIP_AUDIT_VULNERABILITY_PATTERN = (
    r"^Found \d+ known vulnerabilit\w{1,3} in \d+ package\w?$"
)


@dataclass
class PipAuditException(Exception):
    return_code: int
    stdout: str
    stderr: str

    def __init__(self, subprocess_output: subprocess.CompletedProcess) -> None:
        self.return_code = subprocess_output.returncode
        self.stdout = subprocess_output.stdout
        self.stderr = subprocess_output.stderr


class VulnerabilitySource(str, Enum):
    CVE = "CVE"
    CWE = "CWE"
    GHSA = "GHSA"
    PYSEC = "PYSEC"

    @classmethod
    def from_prefix(cls, name: str) -> VulnerabilitySource | None:
        for el in cls:
            if name.upper().startswith(el.value):
                return el
        return None

    def get_link(self, package: str, vuln_id: str) -> str:
        if self == VulnerabilitySource.CWE:
            cwe_id = vuln_id.upper().replace(f"{VulnerabilitySource.CWE.value}-", "")
            return f"https://cwe.mitre.org/data/definitions/{cwe_id}.html"

        map_link = {
            VulnerabilitySource.CVE: "https://nvd.nist.gov/vuln/detail/{vuln_id}",
            VulnerabilitySource.GHSA: "https://github.com/advisories/{vuln_id}",
            VulnerabilitySource.PYSEC: "https://github.com/pypa/advisory-database/blob/main/vulns/{package}/{vuln_id}.yaml",
        }
        return map_link[self].format(package=package, vuln_id=vuln_id)


class Vulnerability(BaseModel):
    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    package: Package
    id: str
    aliases: list[str]
    fix_versions: list[str]
    description: str

    @classmethod
    def from_audit_entry(
        cls, package_name: str, version: str, vuln_entry: dict[str, Any]
    ) -> Vulnerability:
        """
        Create a Vulnerability from a pip-audit vulnerability entry
        """
        return cls(
            package=Package(name=package_name, version=version),
            id=vuln_entry["id"],
            aliases=vuln_entry["aliases"],
            fix_versions=vuln_entry["fix_versions"],
            description=vuln_entry["description"],
        )

    @property
    def references(self) -> list[str]:
        return sorted([self.id] + self.aliases)

    @property
    def reference_links(self) -> tuple[str, ...]:
        return tuple(
            source.get_link(package=self.package.name, vuln_id=reference)
            for reference in self.references
            if (source := VulnerabilitySource.from_prefix(reference.upper()))
        )

    @property
    def security_issue_entry(self) -> dict[str, str | list[str] | tuple[str, ...]]:
        return {
            "name": self.package.name,
            "version": str(self.package.version),
            "refs": self.references,
            "description": self.description,
            "coordinates": self.package.coordinates,
            "references": self.reference_links,
        }

    @property
    def vulnerability_id(self) -> str | None:
        """
        Ensure a consistent way of identifying a vulnerability for string generation.
        """
        for ref in self.references:
            ref_upper = ref.upper()
            if ref_upper.startswith(VulnerabilitySource.CVE.value):
                return ref
            if ref_upper.startswith(VulnerabilitySource.GHSA.value):
                return ref
            if ref_upper.startswith(VulnerabilitySource.PYSEC.value):
                return ref
        return self.references[0]

    @property
    def subsection_for_changelog_summary(self) -> str:
        """
        Create a subsection to be included in the Summary section of a versioned changelog.
        """
        links_join = "\n* ".join(sorted(self.reference_links))
        references_subsection = f"\n#### References:\n\n* {links_join}\n\n "
        subsection = f"### {self.vulnerability_id} in {self.package.coordinates}\n\n{self.description}\n{references_subsection}"
        return cleandoc(subsection.strip())


def audit_poetry_files(working_directory: Path) -> str:
    """
    Audit the `pyproject.toml` and `poetry.lock` files

    pip-audit evaluates installed packages. This is to provide
    additional security-related information beyond seeing if a given package
    has a known vulnerability. Thus, to audit our `pyproject.toml` and
    `poetry.lock` files without altering a locally sourced poetry environment,
    this function first exports the locked packages to a requirements.txt file.
    Then, pip-audit evaluates the requirements.txt by installing them to a virtualenv
    and then inspecting the dependencies.
    """

    requirements_txt = "requirements.txt"
    output = subprocess.run(
        ["poetry", "export", "--format=requirements.txt"],
        capture_output=True,
        text=True,
        cwd=working_directory,
    )  # nosec
    if output.returncode != 0:
        raise PipAuditException(subprocess_output=output)

    with tempfile.TemporaryDirectory() as path:
        tmpdir = Path(path)
        (tmpdir / requirements_txt).write_text(output.stdout)

        command = ["pip-audit", "-r", requirements_txt, "-f", "json"]
        output = subprocess.run(
            command,
            capture_output=True,
            text=True,
            cwd=tmpdir,
        )  # nosec

    if output.returncode != 0:
        # pip-audit does not distinguish between 1) finding vulnerabilities
        # and 2) other errors performing the pip-audit (i.e. malformed file);
        # they both map to returncode = 1, so we have our own logic to raise errors
        # for the case of 2) and not 1).
        if not search(PIP_AUDIT_VULNERABILITY_PATTERN, output.stderr.strip()):
            raise PipAuditException(subprocess_output=output)
    return output.stdout


class Vulnerabilities(BaseModel):
    vulnerabilities: list[Vulnerability]

    @classmethod
    def load_from_pip_audit(cls, working_directory: Path) -> Vulnerabilities:
        """
        Convert the pip-audit JSON output into a Vulnerabilities model

        The output from pip-audit is a JSON, which as a dictionary looks like:
        >>> audit_dict = {"dependencies": [
        ... {"name": "alabaster", "version": "0.7.16", "vulns": []},
        ... {"name": "cryptography", "version": "43.0.3", "vulns":
        ... [{"id": "GHSA-79v4-65xg-pq4g", "fix_versions": ["44.0.1"],
        ... "aliases": ["CVE-2024-12797"],
        ... "description": "pyca/cryptography\'s wheels..."}, ...]}]}
        """
        audit_json = audit_poetry_files(working_directory)
        audit_dict = json.loads(audit_json)

        vulnerabilities = []
        for entry in audit_dict["dependencies"]:
            for vuln_entry in entry["vulns"]:
                vulnerabilities.append(
                    Vulnerability.from_audit_entry(
                        package_name=entry["name"],
                        version=entry["version"],
                        vuln_entry=vuln_entry,
                    )
                )
        return Vulnerabilities(vulnerabilities=vulnerabilities)

    @property
    def security_issue_dict(self) -> list[dict[str, str | list[str] | tuple[str, ...]]]:
        return [
            vulnerability.security_issue_entry for vulnerability in self.vulnerabilities
        ]


def get_vulnerabilities(working_directory: Path) -> list[Vulnerability]:
    return Vulnerabilities.load_from_pip_audit(
        working_directory=working_directory
    ).vulnerabilities


def get_vulnerabilities_from_latest_tag(root_path: Path):
    with poetry_files_from_latest_tag(root_path=root_path) as tmp_dir:
        return get_vulnerabilities(working_directory=tmp_dir)



from __future__ import annotations

import tempfile
from collections.abc import Generator
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Annotated,
    Final,
    NewType,
)

from packaging.version import Version
from pydantic import (
    AfterValidator,
    BaseModel,
    ConfigDict,
)

from exasol.toolbox.util.git import Git

NormalizedPackageStr = NewType("NormalizedPackageStr", str)

VERSION_TYPE = Annotated[str, AfterValidator(lambda v: Version(v))]


def normalize_package_name(package_name: str) -> NormalizedPackageStr:
    return NormalizedPackageStr(package_name.lower().replace("_", "-"))


def create_package_coordinates(package_name: str, version: str | Version) -> str:
    """
    Create a naming convention for combining a package name and its version
    """
    return f"{package_name}:{version}"


class Package(BaseModel):
    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    name: str
    version: VERSION_TYPE

    @property
    def coordinates(self) -> str:
        return create_package_coordinates(package_name=self.name, version=self.version)

    @property
    def normalized_name(self) -> NormalizedPackageStr:
        return normalize_package_name(self.name)


@dataclass(frozen=True)
class PoetryFiles:
    pyproject_toml: Final[str] = "pyproject.toml"
    poetry_lock: Final[str] = "poetry.lock"

    @property
    def files(self) -> tuple[str, ...]:
        return tuple(self.__dict__.values())


@contextmanager
def poetry_files_from_latest_tag(root_path: Path) -> Generator[Path]:
    """Context manager to set up a temporary directory with poetry files from the latest tag"""
    latest_tag = Git.get_latest_tag()
    path = root_path.relative_to(Git.toplevel())
    with tempfile.TemporaryDirectory() as tmpdir_str:
        tmp_dir = Path(tmpdir_str)
        for file in PoetryFiles().files:
            Git.checkout(latest_tag, path / file, tmp_dir / file)
        yield tmp_dir



from __future__ import annotations

import subprocess
import tempfile
from collections import OrderedDict
from dataclasses import dataclass
from inspect import cleandoc
from json import loads

from pydantic import field_validator

from exasol.toolbox.util.dependencies.shared_models import (
    NormalizedPackageStr,
    Package,
)

LICENSE_MAPPING_TO_ABBREVIATION = {
    "BSD License": "BSD",
    "MIT License": "MIT",
    "The Unlicensed (Unlicensed)": "Unlicensed",
    "Mozilla Public License 2.0 (MPL 2.0)": "MPLv2",
    "GNU General Public License (GPL)": "GPL",
    "GNU Lesser General Public License v2 (LGPLv2)": "LGPLv2",
    "GNU General Public License v2 (GPLv2)": "GPLv2",
    "GNU General Public License v2 or later (GPLv2+)": "GPLv2+",
    "GNU General Public License v3 (GPLv3)": "GPLv3",
    "Apache Software License": "Apache",
}

LICENSE_MAPPING_TO_URL = {
    "GPLv1": "https://www.gnu.org/licenses/old-licenses/gpl-1.0.html",
    "GPLv2": "https://www.gnu.org/licenses/old-licenses/gpl-2.0.html",
    "LGPLv2": "https://www.gnu.org/licenses/old-licenses/lgpl-2.0.html",
    "GPLv3": "https://www.gnu.org/licenses/gpl-3.0.html",
    "LGPLv3": "https://www.gnu.org/licenses/lgpl-3.0.html",
    "Apache": "https://www.apache.org/licenses/LICENSE-2.0",
    "MIT": "https://mit-license.org/",
    "BSD": "https://opensource.org/license/bsd-3-clause",
}


class PackageLicense(Package):
    package_link: str | None
    license: str

    @field_validator("package_link", mode="before")
    def map_unknown_to_none(cls, v) -> str | None:
        if v == "UNKNOWN":
            return None
        return v

    @field_validator("license", mode="before")
    def map_to_normalized_values(cls, v) -> str | None:
        return _normalize(v)

    @property
    def license_link(self) -> str | None:
        return LICENSE_MAPPING_TO_URL.get(self.license, None)


def _normalize(_license: str) -> str:
    def is_multi_license(l: str) -> bool:
        return ";" in l

    def select_most_restrictive(licenses: list[str]) -> str:
        lic = "Unknown"
        _mapping = {
            "Unknown": -1,
            "Unlicensed": 0,
            "BSD": 1,
            "MIT": 2,
            "MPLv2": 3,
            "LGPLv2": 4,
            "GPLv2": 5,
            "GPLv3": 6,
        }
        for l in licenses:
            if l in _mapping:
                if _mapping[l] > _mapping[lic]:
                    lic = l
            else:
                return "<br>".join(licenses)
        return lic

    if is_multi_license(_license):
        items = []
        for item in _license.split(";"):
            item = str(item).strip()
            items.append(LICENSE_MAPPING_TO_ABBREVIATION.get(item, item))
        return select_most_restrictive(items)

    return LICENSE_MAPPING_TO_ABBREVIATION.get(_license, _license)


def _packages_from_json(json: str) -> dict[NormalizedPackageStr, PackageLicense]:
    packages = loads(json)
    return {
        package_license.normalized_name: package_license
        for package in packages
        if (
            package_license := PackageLicense(
                name=package["Name"],
                package_link=package["URL"],
                version=package["Version"],
                license=package["License"],
            )
        )
    }


def get_licenses() -> dict[NormalizedPackageStr, PackageLicense]:
    with tempfile.NamedTemporaryFile() as file:
        subprocess.run(
            [
                "pip-licenses",
                "--format=json",
                "--output-file=" + file.name,
                "--with-system",
                "--with-urls",
            ],
            capture_output=True,
            check=True,
        )
        return _packages_from_json(file.read().decode())


@dataclass(frozen=True)
class PackageLicenseReport:
    dependencies: OrderedDict[str, dict[NormalizedPackageStr, Package]]
    licenses: dict[NormalizedPackageStr, PackageLicense]

    @staticmethod
    def _format_group_table_header(group: str) -> str:
        return cleandoc(
            f"""## `{group}` Dependencies
            |Package|Version|License|
            |---|---|---|
            """
        )

    def _format_group_table(
        self, group: str, group_package_names: set[NormalizedPackageStr]
    ) -> str:
        group_header = self._format_group_table_header(group=group)

        rows = []
        for package_name in sorted(group_package_names):
            if license_info := self.licenses.get(package_name):
                rows.append(self._format_table_row(license_info=license_info))

        return f"""{group_header}\n{''.join(rows)}\n"""

    @staticmethod
    def _format_table_row(license_info: PackageLicense) -> str:
        row_package = f"{license_info.name}"
        if license_info.package_link:
            row_package = f"[{license_info.name}]({license_info.package_link})"

        row_license = f"{license_info.license}"
        if license_info.license_link:
            row_license = f"[{license_info.license}]({license_info.license_link})"

        return f"|{row_package}|{license_info.version}|{row_license}|\n"

    def to_markdown(self) -> str:
        rows = []
        for group in self.dependencies:
            group_package_names = set(self.dependencies[group].keys())
            rows.append(
                self._format_group_table(
                    group=group, group_package_names=group_package_names
                )
            )
        return cleandoc(f"""# Dependencies\n\n{''.join(rows)}""")
